import asyncio
import sys

# rmtree msg on win32
import warnings

# import os
import argparse


from pathlib import Path
import hashlib

# import urllib
import shutil
from datetime import datetime

from .__init__ import __version__

import pygbag

from . import pack
from . import web


devmode = "--dev" in sys.argv

DEFAULT_SCRIPT = "main.py"
CACHE_ROOT = Path("build")
CACHE_PATH = CACHE_ROOT / "web-cache"
CACHE_VERSION = CACHE_ROOT / "version.txt"
CACHE_APP = CACHE_ROOT / "web"

cdn_dot = __version__.split(".")
cdn_dot.pop()
cdn_version = ".".join(cdn_dot)
del cdn_dot

AUTO_REBUILD = True

if devmode:
    sys.argv.remove("--dev")
    DEFAULT_PORT = 8666
    DEFAULT_CDN = f"http://localhost:8000/archives/{cdn_version}/"
    DEFAULT_TMPL = "static/wip.tmpl"
    print(
        f"""

*************************************
    DEV MODE

    {DEFAULT_CDN=}
    {DEFAULT_PORT=}
    {DEFAULT_TMPL=}
*************************************
"""
    )

else:
    DEFAULT_CDN = f"https://pygame-web.github.io/archives/{cdn_version}/"
    DEFAULT_PORT = 8000
    DEFAULT_TMPL = "default.tmpl"


def set_args(program):
    global DEFAULT_SCRIPT
    import sys
    from pathlib import Path

    required = []

    patharg = Path(program).resolve()
    if patharg.is_file():
        mainscript = patharg.name
        app_folder = patharg.parent
    else:
        app_folder = patharg.resolve()
        mainscript = DEFAULT_SCRIPT

    sys.path.insert(0, str(app_folder))

    if patharg.suffix == "pyw":
        required.append("79: Error, no .pyw allowed use .py for python script")

    script_path = app_folder / mainscript

    if not script_path.is_file():
        required.append(f"83: Error, no main.py {script_path} found in folder")

    if not app_folder.is_dir() or patharg.as_posix().endswith("/pygbag/__main__.py"):
        required.append("89: Error, Last argument must be a valid app top level directory, or the main.py python script")

    if sys.version_info < (3, 8):
        # zip deflate compression level 3.7
        # https://docs.python.org/3.11/library/shutil.html#shutil.copytree dirs_exist_ok = 3.8
        required.append("84: Error, pygbag requires CPython version >= 3.8")

    if len(required):
        while len(required):
            print(required.pop())
        print("89: missing requirement(s)")
        sys.exit(89)

    return app_folder, mainscript


def cache_check(app_folder, devmode=False):
    global CACHE_PATH, CACHE_APP, __version__

    version_file = app_folder / CACHE_VERSION

    clear_cache = False

    # always clear the cache in devmode, because cache source is local and changes a lot
    if devmode:
        print("103: DEVMODE: clearing cache")
        clear_cache = True
    elif version_file.is_file():
        try:
            with open(version_file, "r") as file:
                cache_ver = file.read()
                if cache_ver != __version__:
                    print(f"115: cache {cache_ver} mismatch, want {__version__}, cleaning ...")
                    clear_cache = True
        except:
            # something's wrong in cache structure, try clean it up
            clear_cache = True
    else:
        clear_cache = True

    cache_root = app_folder.joinpath(CACHE_ROOT)
    cache_dir = app_folder / CACHE_PATH
    build_dir = app_folder / CACHE_APP

    def make_cache_dirs():
        nonlocal cache_root, cache_dir

        cache_root.mkdir(exist_ok=True)
        build_dir.mkdir(exist_ok=True)
        cache_dir.mkdir(exist_ok=True)

    if clear_cache:
        win32 = sys.platform == "win32"
        if shutil.rmtree.avoids_symlink_attacks or win32:
            if cache_dir.is_dir():
                if win32:
                    warnings.warn("clear cache : rmtree is not safe on that system (win32)")
                shutil.rmtree(cache_dir.as_posix())
        else:
            print(
                "115: cannot clear cache : rmtree is not safe on that system",
                file=sys.stderr,
            )
            raise SystemExit(115)

        # rebuild
        make_cache_dirs()

        with open(version_file, "w") as file:
            file.write(__version__)

    return build_dir, cache_dir


async def main_run(app_folder, mainscript, cdn=DEFAULT_CDN):
    global DEFAULT_PORT, DEFAULT_SCRIPT, APP_CACHE, required

    DEFAULT_SCRIPT = mainscript or DEFAULT_SCRIPT

    build_dir, cache_dir = cache_check(app_folder, devmode)

    sys.argv.pop()

    parser = argparse.ArgumentParser()

    print(
        "\nServing python files from [%s]\n\nwith no security/performance in mind, i'm just a test tool : don't rely on me"
        % build_dir
    )

    parser.add_argument(
        "--bind",
        default="localhost",
        metavar="ADDRESS",
        help="Specify alternate bind address [default: localhost]",
    )
    parser.add_argument(
        "--directory",
        default=build_dir.as_posix(),
        help="Specify alternative directory [default:%s]" % build_dir,
    )

    parser.add_argument(
        "--PYBUILD",
        default="3.11",
        help="Specify python version [default:%s]" % "3.11",
    )
    parser.add_argument(
        "--app_name",
        default=app_folder.name,
        help="Specify user facing name of application [default:%s]" % app_folder.name,
    )

    parser.add_argument(
        "--ume_block",
        default=1,
        help="Specify wait for user media engagement before running [default:%s]" % 1,
    )

    parser.add_argument(
        "--can_close",
        default=0,
        help="Specify if window will ask confirmation for closing [default:%s]" % 0,
    )

    parser.add_argument("--cache", default=cache_dir.as_posix(), help="md5 based url cache directory")

    parser.add_argument(
        "--package",
        default=f"web.pygame.{app_folder.name}-{int(datetime.timestamp(datetime.now()))}",
        help="package name, better make it unique",
    )

    parser.add_argument("--title", default="", help="App nice looking name")

    parser.add_argument(
        "--version",
        default=__version__,
        help="override prebuilt version path [default:%s]" % __version__,
    )

    parser.add_argument("--build", action="store_true", help="build only, do not run test server")

    parser.add_argument(
        "--html",
        action="store_true",
        help="build as html with embedded assets (pygame-script)",
    )

    parser.add_argument("--no_opt", action="store_true", help="turn off assets optimizer")

    parser.add_argument("--archive", action="store_true", help="make build/web.zip archive for itch.io")

    #    parser.add_argument(
    #        "--main",
    #        default=DEFAULT_SCRIPT,
    #        help="Specify main script" "[default:%s]" % DEFAULT_SCRIPT,
    #    )

    parser.add_argument(
        "--icon",
        default="favicon.png",
        help="icon png file 32x32 min should be favicon.png",
    )

    parser.add_argument(
        "--cdn",
        default=cdn,
        help="web site to cache locally [default:%s]" % cdn,
    )

    parser.add_argument(
        "--template",
        default=DEFAULT_TMPL,
        help="index.html template [default:%s]" % DEFAULT_TMPL,
    )

    parser.add_argument("--ssl", default=False, help="enable ssl with server.pem and key.pem")

    parser.add_argument(
        "--port",
        action="store",
        default=DEFAULT_PORT,
        type=int,
        nargs="?",
        help="Specify alternate port [default: 8000]",
    )

    args = parser.parse_args()

    app_name = app_folder.name.lower().replace(" ", ".")

    archfile = build_dir.joinpath(f"{app_name}.apk")

    if archfile.is_file():
        archfile.unlink()

    print(
        f"""

SUMMARY
________________________

# the app folder
app_folder={app_folder}

# artefacts directory
build_dir={build_dir}

# the window title and icon name
app_name={app_name}

# package name, better make it unique
package={args.package}

# icons:  96x96 for desktop, 16x16 for web
icon={args.icon}

# js/wasm provider
cdn={args.cdn}

now packing application ....

"""
    )

    CC = {
        "cdn": args.cdn,
        "proxy": f"http://{args.bind}:{args.port}/",
        "xtermjs": "1",
        "ume_block": args.ume_block,
        "can_close": args.can_close,
        "archive": app_name,
        "autorun": "0",
        "authors": "pgw",
        "icon": args.icon,
        "title": (args.title or app_name),
        "directory": app_name,
        "spdx": "cookiecutter.spdx",
        "version": __version__,
        "PYBUILD": args.PYBUILD,
    }

    pygbag.config = CC

    await pack.archive(f"{app_name}.apk", app_folder, build_dir)

    def cache_file(remote_url, suffix):
        nonlocal cache_dir
        cache = hashlib.md5(remote_url.encode()).hexdigest()
        cached = cache_dir.joinpath(cache + "." + suffix)
        return cached

    # get local or online template in order
    # _______________________________________

    template_file = Path(args.template)

    if template_file.is_file():
        print(
            f"""
        building from local template {args.template}
        result files will be in {build_dir}
"""
        )
    else:
        tmpl_url = f"{args.cdn}{args.template}"
        tmpl = cache_file(tmpl_url, "tmpl")
        if tmpl.is_file():
            print(
                f"""
    building from local cached template {args.cdn}{args.template}
    cached at {tmpl}
    result files will be in {build_dir}
"""
            )
            template_file = tmpl
        else:
            print(
                f"""
    caching template {args.cdn}{args.template}
    cached locally at {tmpl}
    result files will be in {build_dir}
"""
            )

            try:
                template_file, headers = web.get(tmpl_url, tmpl)
            except Exception as e:
                print(e)
                print(f"CDN {args.cdn} is not responding : not running test server")
                args.build = True

    if app_folder.joinpath("static").is_dir():
        print(
            f"""
        copying static files to webroot {build_dir}
"""
        )
        # dirs_exist_ok = 3.8
        shutil.copytree(app_folder.joinpath("static"), build_dir, dirs_exist_ok=True)

    # get local or online favicon in order
    # _______________________________________
    icon_file = Path(args.icon)
    if not icon_file.is_file():
        icon_url = f"{args.cdn}{args.icon}"
        icon_file = cache_file(icon_url, "png")

        if icon_file.is_file():
            print(
                f"""
    using icon from local cached file {icon_url}
    cached at {icon_file}"""
            )

        else:
            try:
                icon_file, headers = web.get(icon_url, icon_file)
                print(
                    f"""
        caching icon {icon_url}
        cached locally at {icon_file}
        """
                )
            except Exception as e:
                sys.print_exception(e)
                print(f"{icon_url} caching error :", e)

    if icon_file.is_file():
        shutil.copyfile(icon_file, build_dir.joinpath("favicon.png"))
    else:
        print(f"error: cannot find {icon_file=}")

    if template_file.is_file():
        with template_file.open("r", encoding="utf-8") as source:
            with open(build_dir.joinpath("index.html").resolve(), "w", encoding="utf-8") as target:
                # while ( line := source.readline()):
                while True:
                    line = source.readline()
                    if not line:
                        break
                    for k, v in CC.items():
                        line = line.replace("{{cookiecutter." + k + "}}", str(v))

                    target.write(line)

        # files should be all ready and tested now
        # except on CDN error on first test, but you did test didn't you ?
        if args.archive:
            print(
                f"""
    only building a web archive suitable for itch.io
    archive file will be :
        {build_dir}.zip

"""
            )

            await pack.web_archive(f"{app_name}.apk", build_dir)
            return

        elif not args.build:

            from . import testserver

            testserver.run_code_server(args, CC)

        else:
            print(
                f"""
    build only requested, not running testserver, files ready here :

build_dir = {build_dir}

"""
            )
    else:
        print(args.template, "is not a valid template")


def main():
    app_folder, mainscript = set_args(sys.argv[-1])

    # sim does not use cache.
    if "--sim" in sys.argv:
        print(f"To use simulator launch with : {sys.executable} -m pygbag {' '.join(sys.argv[1:])}")
        return 1
    else:
        asyncio.run(main_run(app_folder, mainscript))
    return 0



import os
import os.path

import urllib
import urllib.request

import time

from pathlib import Path

try:
    import ssl
except:

    class ssl:
        SSLCertVerificationError = urllib.error.URLError


import sys

# https://stackoverflow.com/questions/42098126/mac-osx-python-ssl-sslerror-ssl-certificate-verify-failed-certificate-verify
def fixcert():
    wd = os.getcwd()
    import stat
    import subprocess

    STAT_0o775 = (
        stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH
    )

    openssl_dir, openssl_cafile = os.path.split(ssl.get_default_verify_paths().openssl_cafile)

    try:
        import certifi
    except:
        print("running : pip install --user --upgrade certifi")
        subprocess.check_call(
            [
                sys.executable,
                "-E",
                "-s",
                "-m",
                "pip",
                "install",
                "--user",
                "--upgrade",
                "certifi",
            ]
        )

    import certifi

    # change working directory to the default SSL directory
    os.chdir(openssl_dir)
    relpath_to_certifi_cafile = os.path.relpath(certifi.where())
    print(" -- removing any existing file or link")
    try:
        os.remove(openssl_cafile)
    except FileNotFoundError:
        pass
    print(" -- creating symlink to certifi certificate bundle")
    os.symlink(relpath_to_certifi_cafile, openssl_cafile)
    print(" -- setting permissions")
    os.chmod(openssl_cafile, STAT_0o775)
    os.chdir(wd)
    print(" -- update complete")


def get(url, path):
    error = None
    data_file = None
    try:
        print(f'urllib.request.urlretrieve("{url}", "{path}")')
        while data_file is None:
            try:
                data_file, header = urllib.request.urlretrieve(url, path)

            except urllib.error.HTTPError as e:
                error = e

            except urllib.error.URLError as e:
                error = e

            except ssl.SSLCertVerificationError:
                print("Trying to fix certificate error")
                fixcert()
                continue

            if data_file is not None:
                break

            if error:
                print(f"WARNING: web.get({url}) : {error}")

            time.sleep(5)
            print("retrying in 5 seconds")

    finally:
        if data_file is not None:
            return Path(data_file), header
        # this is normal in dev mode for favicon and templates because
        # proxy is not yet started.
        print(f"NO DATA RECEIVED FOR {url}")
        raise Exception(f"cannot cache {url} to {path}")


if __name__ == "__main__":
    fixcert()



import asyncio

asyncrun = asyncio.run

import sys
from .__init__ import __version__

print(f" *pygbag {__version__}*")

from pathlib import Path
from .app import main_run, set_args


async def import_site():
    import sys
    from pathlib import Path

    app_folder, mainscript = set_args(sys.argv[-1])

    if ("--sim" not in sys.argv) and ("--piny" not in sys.argv):
        # run pygbag build/server
        await main_run(app_folder, mainscript)
        return True

    # or run as a native simulator

    mod_dir = Path(__file__).parent
    support = mod_dir / "support"

    print(" - simulator -", mod_dir, support)

    import sys, os, builtins

    sys.path.insert(0, str(support / "cross"))

    #    try:
    #        import pymunk4 as pymunk
    #
    #        sys.modules["pymunk"] = pymunk
    #    except:
    #        print("pymunk4 was not build for simulator")

    # need them earlier than aio

    def pdb(*argv):
        print(*argv, file=sys.__stderr__)

    builtins.pdb = pdb

    def print_exception(e, out=sys.stderr, **kw):
        kw["file"] = out
        traceback.print_exc(**kw)

    # cannot fake a cpu __WASM__ will be False

    # but fake the platform AND the module
    sys.platform = "emscripten"

    class fake_EventTarget:
        clients = {}
        events = []

        async def process(self):
            ...

    # et = EventTarget()
    class __EMSCRIPTEN__(object):
        EventTarget = fake_EventTarget()
        ticks = 0

        def __init__(self):
            import platform

            self.platform = platform

        def __getattribute__(self, name):
            try:
                return object.__getattribute__(self, name)
            except:
                return object.__getattribute__(self.platform, name)

        @classmethod
        def PyConfig_InitPythonConfig(*argv):
            pass

        @classmethod
        def init_platform(*argv):
            pass

        @classmethod
        def flush(cls):
            sys.stdout.flush()
            sys.stderr.flush()

        @classmethod
        def system(cls):
            return "Linux"

        @classmethod
        def no_op(cls, *argv, **kw):
            ...

        run = no_op

        set_ps1 = no_op

        prompt = no_op

        is_browser = False

        js = pdb
        run_script = pdb

    __EMSCRIPTEN__ = __EMSCRIPTEN__()

    sys.modules["__EMSCRIPTEN__"] = __EMSCRIPTEN__
    sys.modules["embed"] = __EMSCRIPTEN__

    with open(support / "pythonrc.py", "r") as file:
        exec(file.read(), globals(), globals())

    import zipfile
    import aio.toplevel
    import ast
    from pathlib import Path

    class TopLevel_async_handler(aio.toplevel.AsyncInteractiveConsole):

        HTML_MARK = '""" # BEGIN -->'

        @classmethod
        async def async_repos(cls):
            abitag = f"cp{sys.version_info.major}{sys.version_info.minor}"
            print(f"{abitag=}")

        @classmethod
        async def async_imports(cls, callback, *wanted, **kw):
            ...

        @classmethod
        def list_imports(cls, code=None, file=None):
            return []

        def eval(self, source):
            for count, line in enumerate(source.split("\n")):
                if not count:
                    if line.startswith("<"):
                        self.buffer.append(f"#{line}")
                        continue
                self.buffer.append(line)

            if count:
                self.line = None
                self.buffer.insert(0, "#")
            print(f"178: {count} lines queued for async eval")

    # start async top level machinery and add a console.
    await TopLevel_async_handler.start_toplevel(platform.shell, console=True)
    ns = vars(__import__(__name__))

    ns["TopLevel_async_handler"] = TopLevel_async_handler

    sourcefile = sys.argv[-1]

    __import__(__name__).__file__ = str(sourcefile)

    if "--piny" in sys.argv:
        from . import mutator

        mutator.transform_file(sourcefile, f"{sourcefile[:-3]}.pn")

    else:

        import aio.clock

        # asyncio.create_task( aio.clock.loop() )
        aio.clock.start(x=80)

        print(__name__, "sim repl ready for", __file__)

        await shell.runpy(__file__)
        shell.interactive()

        # if you don't reach that step
        # your main.py has an infinite sync loop somewhere !
        print(f"{platform.is_browser=}, sim ready, press enter to start")

        while not aio.exit:
            await aio.sleep(0.016)
        print(__name__, "sim terminated")


if __name__ == "__main__":
    asyncio.run(import_site())


#



from pathlib import Path

import pygbag


def stringify(blob):
    c = 0
    for b in blob:
        yield chr(int(b) + 248)
        c += 1
        if c > 78:
            yield "\n"
            c = 0


def dump_fs(html, target_folder, packlist):
    html.write(
        f"""PYGBAG_FS={len(packlist)}
# fmt: off
__import__('os').chdir(__import__('tempfile').gettempdir())
def fs_decode(fsname, o248):
    from pathlib import Path
    filename = Path.cwd() / fsname
    if not filename.is_file():
        filename.parent.mkdir(parents=True, exist_ok=True)
        with open(fsname,"wb") as fs:
            for input in o248.split("\\n"):
                if not input: continue
                fs.write(bytes([ord(c) - 248 for c in input]))
"""
    )

    for topack in packlist:
        if topack == "/main.py":
            continue

        vfs_name = topack[1:].replace("-pygbag.", ".")

        src_name = target_folder / topack[1:]

        sum = str(src_name.stat().st_size)

        if topack.lower().endswith(".py"):
            html.write(
                f'''
with open("{vfs_name}","w") as fs:fs.write("""\\
{open(src_name, "r").read()}""")
'''
            )
            html.write("\n")

        else:
            html.write(f"\nfs_decode('{vfs_name}','''\n")
            for text in stringify(open(src_name, "rb").read()):
                html.write(text)
            html.write("''')\n")

    html.write("\n# fmt:on\ndel fs_decode, PYGBAG_FS\n")


def make_header(html, line):
    if line and line[0] == "<":
        pass  # already a script shebang
    else:
        if line.find("pythons.js") > 0:
            # found script directives after #!
            SCRIPT = line[2:].strip()
        else:
            SCRIPT = f' src="{pygbag.config["cdn"]}pythons.js"'
            SCRIPT += ' data-os="vtx,fs,gui"'

        line = (
            f"""
<html>
<head>
<meta charset="utf-8">
</head>
<script {SCRIPT} type=module id="__main__"  async defer>
#<!--
""".replace(
                "\n", ""
            )
            .replace("  ", " ")
            .strip()
        )
    print(line, end="\n", file=html)


def html_embed(target_folder, packlist: list, htmlfile: str):
    print("HTML:", htmlfile)
    RUNPY = "asyncio.run(main())"
    SKIP = False
    MAX = 0
    with open(htmlfile, "w+", encoding="utf-8") as html:
        for topack in packlist:

            if topack == "/main.py":
                for lnum, line in enumerate(open(target_folder / topack[1:], "r", encoding="utf-8").readlines()):
                    if line.rstrip().startswith("asyncio.run"):
                        RUNPY = line
                        MAX = lnum
                        break

                for lnum, line in enumerate(open(target_folder / topack[1:], "r", encoding="utf-8").readlines()):
                    if SKIP:
                        if line.endswith("del fs_decode, PYGBAG_FS"):
                            SKIP = False
                    if line.startswith("PYGBAG_FS="):
                        SKIP = True

                    if SKIP:
                        continue

                    # no msdos
                    line = line.rstrip("\r\n")

                    if not lnum:
                        make_header(html, line)
                        dump_fs(html, target_folder, packlist)
                        continue
                    else:
                        if lnum >= MAX:
                            break

                    print(line, end="\n", file=html)

            break

        print(
            f"""
{RUNPY}
# --></script></html>
""",
            file=html,
        )



# https://emscripten.org/docs/porting/networking.html
# idea : Katie Bell ( https://github.com/katharosada ) on WebAssembly/Python discord


# possible client part : https://github.com/Anorov/PySocks
# this code : Han You : https://github.com/sqd/pysocks5-async


import asyncio
import errno
import ipaddress
import logging
import sys

logging.basicConfig(level=logging.INFO)
from typing import Union

_BUF_LEN = 64

_ACCEPTED_VERSION = b"\x05"
_NO_AUTH = b"\x00"

_NO_SUPPORTED_AUTH_RESP = b"\x05\xff"
_USE_NO_AUTH_RESP = b"\x05\x00"


def _status_code_resp(code):
    return b"\x05" + code + b"\x00\x01\x00\x00\x00\x00\x00\x00"


_GENERAL_FAILURE_RESP = _status_code_resp(b"\x01")
_NOT_SUPPORTED_CMD_RESP = _status_code_resp(b"\x07")
_NOT_SUPPORTED_ADDR_RESP = _status_code_resp(b"\x08")

_ADDR_TYPE_IPV4 = b"\x01"
_ADDR_TYPE_DOMAIN = b"\x03"
_ADDR_TYPE_IPV6 = b"\x04"
_ACCEPTED_ADDR_TYPES = (_ADDR_TYPE_IPV4, _ADDR_TYPE_DOMAIN, _ADDR_TYPE_IPV6)

_CMD_TCP_OPEN = b"\x01"
_CMD_TCP_BIND = b"\x02"
_CMD_UDP_ASSOC = b"\x03"
_ACCEPTED_CMDS = (_CMD_TCP_OPEN, _CMD_TCP_BIND, _CMD_UDP_ASSOC)


class SOCKS5Server:
    def __init__(self, host, port, handler_cls):
        self.host = host
        self.port = port
        self.handler_cls = handler_cls

    async def start_server(self):
        logging.info("Start listening at %s:%s", self.host, self.port)
        await asyncio.start_server(self._handle_conn, host=self.host, port=self.port)

    async def _handle_conn(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        def close_conn(resp):
            writer.write(resp)
            writer.close()

        version = await reader.read(1)
        if version != _ACCEPTED_VERSION:
            return close_conn(_GENERAL_FAILURE_RESP)

        n_methods = (await reader.read(1))[0]
        auth_methods = await reader.read(n_methods)
        if _NO_AUTH not in auth_methods:
            return close_conn(_NO_SUPPORTED_AUTH_RESP)
        else:
            writer.write(_USE_NO_AUTH_RESP)

        try:
            version = await reader.read(1)
            if version != _ACCEPTED_VERSION:
                return close_conn(_GENERAL_FAILURE_RESP)

            cmd = await reader.read(1)
            if cmd not in _ACCEPTED_CMDS:
                return close_conn(_NOT_SUPPORTED_CMD_RESP)

            reserve_zero = await reader.read(1)
            if reserve_zero != b"\x00":
                return close_conn(_GENERAL_FAILURE_RESP)

            addr_type = await reader.read(1)
            if addr_type not in _ACCEPTED_ADDR_TYPES:
                return close_conn(_NOT_SUPPORTED_ADDR_RESP)

            if addr_type == _ADDR_TYPE_IPV4:
                host = ipaddress.IPv4Address((await reader.read(4)))
            elif addr_type == _ADDR_TYPE_DOMAIN:
                domain_len = (await reader.read(1))[0]
                host = (await reader.read(domain_len)).decode("ascii")
            elif addr_type == _ADDR_TYPE_IPV6:
                host = ipaddress.IPv6Address((await reader.read(16)))

            port = int.from_bytes((await reader.read(2)), "big")

            handler: BaseSOCKS5Handler = self.handler_cls(reader, writer, host, port)
        except ConnectionError:
            return close_conn(_GENERAL_FAILURE_RESP)

        if cmd == _CMD_TCP_OPEN:
            await handler.do_TCP_open()
        elif cmd == _CMD_TCP_BIND:
            await handler.do_TCP_bind()
        elif cmd == _CMD_UDP_ASSOC:
            await handler.do_UDP_assoc()


class SOCKS5Status:
    OK = b"\x00"
    GENERAL_FAILURE = b"\x01"
    NOT_ALLOWED = b"\x02"
    NETWORK_UNREACHABLE = b"\x03"
    HOST_UNREACHABLE = b"\x04"
    CONN_REFUSED = b"\x05"
    TTL_EXPIRED = b"\x06"
    PROTO_ERR = b"\x07"
    ADDR_NOT_SUPPORTED = b"\x08"


class BaseSOCKS5Handler:
    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        host: Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address],
        port: int,
    ):
        self.client_reader = reader
        self.client_writer = writer
        self.dest_host = host
        self.dest_port = port
        self.client_host, self.client_port = writer.get_extra_info("peername")

    def dest_host_str(self):
        if isinstance(self.dest_host, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            return self.dest_host.exploded
        elif isinstance(self.dest_host, str):
            return self.dest_host

    async def do_TCP_open(self):
        pass

    async def do_TCP_bind(self):
        pass

    async def do_UDP_assoc(self):
        pass

    def response_status(self, status):
        resp = b"\x05"
        resp += status
        resp += b"\x00"
        if isinstance(self.dest_host, ipaddress.IPv4Address):
            resp += b"\x01" + self.dest_host.packed
        elif isinstance(self.dest_host, str):
            resp += b"\x03" + len(self.dest_host).to_bytes(1, "big") + self.dest_host.encode("ascii")
        elif isinstance(self.dest_host, ipaddress.IPv6Address):
            resp += b"\x04" + self.dest_host.packed
        resp += self.dest_port.to_bytes(2, "big")
        self.client_writer.write(resp)

    def close(self):
        self.client_writer.close()


class SimpleSOCKS5Handler(BaseSOCKS5Handler):
    async def do_TCP_open(self):
        logging.info(
            "TCP OPEN %s:%s -> %s:%s",
            self.client_host,
            self.client_port,
            self.dest_host_str(),
            self.dest_port,
        )
        self.response_status(SOCKS5Status.OK)
        try:
            server_reader, server_writer = await asyncio.open_connection(self.dest_host_str(), self.dest_port)
        except ConnectionRefusedError:
            self.response_status(SOCKS5Status.CONN_REFUSED)
            return self.close()
        except TimeoutError:
            self.response_status(SOCKS5Status.GENERAL_FAILURE)
            return self.close()
        except OSError as e:
            status = {
                errno.ENETUNREACH: SOCKS5Status.NETWORK_UNREACHABLE,
                errno.EHOSTUNREACH: SOCKS5Status.HOST_UNREACHABLE,
            }.get(e.errno, SOCKS5Status.GENERAL_FAILURE)
            self.response_status(status)
            return self.close()
        await SimpleSOCKS5Handler._bridge(self.client_reader, self.client_writer, server_reader, server_writer)

    async def do_TCP_bind(self):
        async def on_conn(server_reader: asyncio.StreamReader, server_writer: asyncio.StreamWriter):
            await self._bridge(self.client_reader, self.client_writer, server_reader, server_writer)
            srv.close()
            logging.info(
                "TCP UNBOUND %s:%s == %s:%s",
                self.client_host,
                self.client_port,
                self.dest_host_str(),
                self.dest_port,
            )

        logging.info(
            "TCP BIND %s:%s == %s:%s",
            self.client_host,
            self.client_port,
            self.dest_host_str(),
            self.dest_port,
        )
        try:
            srv = await asyncio.start_server(on_conn, self.dest_host_str(), self.dest_port)
            self.dest_host, self.dest_port = srv.sockets[0].getsockname()
            logging.info(
                "TCP BOUND %s:%s == %s:%s",
                self.client_host,
                self.client_port,
                self.dest_host_str(),
                self.dest_port,
            )
            self.response_status(SOCKS5Status.OK)
        except OSError as e:
            self.response_status(SOCKS5Status.GENERAL_FAILURE)
            self.close()

    async def do_UDP_assoc(self):
        class UDPClientProto(asyncio.DatagramProtocol):
            handler = self

            def datagram_received(self, data, addr):
                UDPClientProto.handler.client_writer.write(data)

        logging.info(
            "UDP ASOC %s:%s == %s:%s",
            self.client_host,
            self.client_port,
            self.dest_host_str(),
            self.dest_port,
        )

        transport: asyncio.DatagramTransport
        proto: asyncio.DatagramProtocol
        (transport, proto) = await asyncio.get_running_loop().create_datagram_endpoint(
            UDPClientProto, remote_addr=(self.dest_host_str(), self.dest_port)
        )
        self.dest_host, self.dest_port = transport.get_extra_info("sockname")
        self.response_status(SOCKS5Status.OK)

        try:
            await self.client_reader.read()
        except:
            pass
        finally:
            transport.close()

    @staticmethod
    async def _bridge(
        active_reader: asyncio.StreamReader,
        active_writer: asyncio.StreamWriter,
        passive_reader: asyncio.StreamReader,
        passive_writer: asyncio.StreamWriter,
    ):
        """
        Cut the passive reader/writer if the active ones disconnect.
        """

        async def pipe(
            from_: asyncio.StreamReader,
            to: asyncio.StreamWriter,
            cut: asyncio.StreamWriter,
        ):
            try:
                buf = await from_.read(_BUF_LEN)
                while buf:
                    to.write(buf)
                    buf = await from_.read(_BUF_LEN)
            except ConnectionResetError:
                if cut:
                    cut.close()
            finally:
                to.close()

        t1 = asyncio.create_task(pipe(active_reader, passive_writer, passive_writer))
        t2 = asyncio.create_task(pipe(passive_reader, active_writer, None))
        await asyncio.gather(t1, t2)


if __name__ == "__main__":
    try:
        port = int(sys.argv[1])
    except (IndexError, ValueError):
        print(f"Usage: {sys.argv[0]} [port]")
        print("Han You (me@hanyou.dev) 2019")
        sys.exit(1)
    socks5d = SOCKS5Server("localhost", port, SimpleSOCKS5Handler)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(socks5d.start_server())
    loop.run_forever()



from os import walk
from pathlib import Path


class Error(Exception):
    pass


def gather(root: Path, *kw):
    if root.is_file():
        if root.name == "main.py":
            raise Error("project must be a folder or an archive")

    for current, dirnames, filenames in walk(root):
        rel = Path("/").joinpath(Path(current).relative_to(root))

        # print(rel, len(dirnames), len(filenames))
        yield rel, filenames



""" packager+server for pygbag wasm loader """

# last 0.6.6
__version__ = "0.7.0"


# WaPy=>CPython compat

import builtins

try:
    sched_yield
except:
    builtins.sched_yield = lambda: None

import sys, traceback

if not hasattr(sys, "print_exception"):

    def print_exception(e, out=sys.stderr, **kw):
        kw["file"] = out
        traceback.print_exc(**kw)

    sys.print_exception = print_exception


def ESC(*argv):
    for arg in argv:
        sys.__stdout__.write(chr(0x1B))
        sys.__stdout__.write(arg)


def CSI(*argv):
    for arg in argv:
        sys.__stdout__.write(chr(0x1B))
        sys.__stdout__.write("[")
        sys.__stdout__.write(arg)


builtins.ESC = ESC
builtins.CSI = CSI
